import { BaseService } from "../base/BaseService";
/**
 * Service for retrieving events from the Kleros Escrow contract
 */
export class EventService extends BaseService {
    /**
     * Creates a new EventService instance
     * @param config The Kleros Escrow configuration
     * @param provider Optional provider for read operations
     */
    constructor(config, provider) {
        super(config, provider);
    }
    /**
     * Normalizes a block parameter to ensure it's in the correct format for ethers.js
     * @param blockParam The block parameter (number, string, or undefined)
     * @returns The normalized block parameter
     */
    normalizeBlockParam(blockParam) {
        // Empty implementation
        return undefined;
    }
    /**
     * Common extended search functionality for events
     * @param transactionId The transaction ID to search for
     * @param eventName The name of the event for logging
     * @param filter The event filter to use
     * @param mapFunction Function to map events to the desired type
     * @param options Optional block range
     * @returns Array of found events
     */
    async searchEventsWithExtendedRange(transactionId, eventName, filter, mapFunction, options) {
        try {
            // If specific block range is provided, search only in that range
            if ((options === null || options === void 0 ? void 0 : options.fromBlock) || (options === null || options === void 0 ? void 0 : options.toBlock)) {
                console.log(`Searching for ${eventName} events in specified block range`);
                const fromBlock = this.normalizeBlockParam(options === null || options === void 0 ? void 0 : options.fromBlock);
                const toBlock = this.normalizeBlockParam(options === null || options === void 0 ? void 0 : options.toBlock);
                const events = await this.escrowContract.queryFilter(filter, fromBlock, toBlock);
                const relevantEvents = transactionId
                    ? events.filter(e => { var _a, _b; return ((_b = (_a = e.args) === null || _a === void 0 ? void 0 : _a._transactionID) === null || _b === void 0 ? void 0 : _b.toString()) === transactionId; })
                    : events;
                if (relevantEvents.length > 0) {
                    return this.processEvents(relevantEvents, transactionId, mapFunction);
                }
                return [];
            }
            // Get transaction details
            const transaction = await this.escrowContract.transactions(transactionId);
            const lastInteractionTimestamp = transaction.lastInteraction.toNumber();
            const timeoutPayment = transaction.timeoutPayment.toNumber();
            const feeTimeout = await this.escrowContract.feeTimeout();
            // Calculate maximum possible block range for searching backwards
            const AVERAGE_BLOCK_TIME = 13; // seconds
            const maxTimeInSeconds = timeoutPayment + feeTimeout.toNumber();
            const maxBlocksBack = Math.ceil(maxTimeInSeconds / AVERAGE_BLOCK_TIME);
            // Find the block number for lastInteraction timestamp
            const lastInteractionBlock = await this.findBlockForTimestamp(lastInteractionTimestamp);
            // Calculate search range
            const searchEndBlock = lastInteractionBlock; // Search backwards from last interaction
            const searchStartBlock = Math.max(1, lastInteractionBlock - maxBlocksBack - 10); // Add 10 block buffer
            console.log(`Searching backwards from block ${searchEndBlock} to ${searchStartBlock}`);
            const MAX_BLOCK_RANGE = 10000; // RPC limit
            let allEvents = [];
            // If range is larger than RPC limit, break into chunks
            if (searchEndBlock - searchStartBlock > MAX_BLOCK_RANGE) {
                let currentEndBlock = searchEndBlock;
                console.log(`Starting chunk search from end block ${currentEndBlock} to start block ${searchStartBlock}`);
                while (currentEndBlock > searchStartBlock) {
                    const fromBlock = Math.max(searchStartBlock, currentEndBlock - MAX_BLOCK_RANGE + 1);
                    console.log(`Processing chunk: ${fromBlock} to ${currentEndBlock} (size: ${currentEndBlock - fromBlock + 1} blocks)`);
                    const events = await this.escrowContract.queryFilter(filter, fromBlock, currentEndBlock);
                    console.log(`Found ${events.length} events in this chunk`);
                    allEvents = [...allEvents, ...events];
                    if (fromBlock === searchStartBlock) {
                        // We've reached the start block, exit the loop
                        console.log(`Reached start block ${searchStartBlock}, ending search`);
                        break;
                    }
                    // Move to next chunk, ensuring we don't go below searchStartBlock
                    currentEndBlock = fromBlock - 1;
                    console.log(`Moving to next chunk, new end block: ${currentEndBlock}`);
                }
                console.log(`Chunk search complete, found total ${allEvents.length} events`);
            }
            else {
                // Single search if range is within RPC limit
                console.log(`Searching single range ${searchStartBlock} to ${searchEndBlock}`);
                allEvents = await this.escrowContract.queryFilter(filter, searchStartBlock, searchEndBlock);
                console.log(`Found ${allEvents.length} events in single search`);
            }
            // Filter relevant events
            const relevantEvents = transactionId
                ? allEvents.filter(e => { var _a, _b; return ((_b = (_a = e.args) === null || _a === void 0 ? void 0 : _a._transactionID) === null || _b === void 0 ? void 0 : _b.toString()) === transactionId; })
                : allEvents;
            if (relevantEvents.length > 0) {
                return this.processEvents(relevantEvents, transactionId, mapFunction);
            }
            return [];
        }
        catch (error) {
            console.error(`Error in searchEventsWithExtendedRange for ${eventName}:`, error);
            return [];
        }
    }
    // Helper function to find block number for a timestamp
    async findBlockForTimestamp(timestamp) {
        const currentBlock = await this.provider.getBlockNumber();
        const currentBlockData = await this.provider.getBlock(currentBlock);
        if (timestamp >= currentBlockData.timestamp) {
            return currentBlock;
        }
        // Binary search for the block
        let left = 1;
        let right = currentBlock;
        while (left <= right) {
            const mid = Math.floor((left + right) / 2);
            const midBlock = await this.provider.getBlock(mid);
            if (midBlock.timestamp === timestamp) {
                return mid;
            }
            else if (midBlock.timestamp < timestamp) {
                left = mid + 1;
            }
            else {
                right = mid - 1;
            }
        }
        return right; // Return the closest block before the timestamp
    }
    /**
     * Gets all events for a specific transaction
     * @param transactionId The ID of the transaction
     * @param options Optional filter options including fromBlock and toBlock
     * @returns All events related to the transaction
     */
    async getEventsForTransaction(transactionId, options) {
        try {
            const [metaEvidenceEvents, paymentEvents, hasToPayFeeEvents, disputeEvents, evidenceEvents, rulingEvents] = await Promise.all([
                this.getMetaEvidenceEvents(transactionId, options),
                this.getPaymentEvents(transactionId, options),
                this.getHasToPayFeeEvents(transactionId, options),
                this.getDisputeEvents(transactionId, options),
                this.getEvidenceEvents(transactionId, options),
                this.getRulingEvents(transactionId, options)
            ]);
            // Combine all events and sort by block number
            const allEvents = [
                ...metaEvidenceEvents,
                ...paymentEvents,
                ...hasToPayFeeEvents,
                ...disputeEvents,
                ...evidenceEvents,
                ...rulingEvents
            ].sort((a, b) => a.blockNumber - b.blockNumber);
            return allEvents;
        }
        catch (error) {
            console.error("Error in getEventsForTransaction:", error);
            return [];
        }
    }
    /**
     * Gets MetaEvidence events for a transaction with extended search capability
     * @param transactionId The ID of the transaction
     * @param options Optional filter options including fromBlock and toBlock
     * @returns Array of MetaEvidence events
     */
    async getMetaEvidenceEvents(transactionId, options) {
        return this.searchEventsWithExtendedRange(transactionId, "MetaEvidence", this.escrowContract.filters.MetaEvidence(), (event, blockData, txId) => {
            var _a, _b;
            return ({
                transactionId: txId,
                blockNumber: event.blockNumber,
                transactionHash: event.transactionHash,
                timestamp: (blockData === null || blockData === void 0 ? void 0 : blockData.timestamp) || 0,
                metaEvidenceId: (_a = event.args) === null || _a === void 0 ? void 0 : _a._metaEvidenceID.toString(),
                evidence: (_b = event.args) === null || _b === void 0 ? void 0 : _b._evidence
            });
        }, options);
    }
    /**
     * Process raw events into structured format
     * @param events Raw ethers.js events
     * @param transactionId The transaction ID
     * @param mapFunction Optional function to map the event to a specific type
     * @returns Processed events
     */
    async processEvents(events, transactionId, mapFunction) {
        const results = [];
        // Default mapping function for MetaEvidence events if none provided
        const defaultMapFunction = (event, blockData, txId) => {
            var _a, _b;
            return ({
                transactionId: txId,
                blockNumber: event.blockNumber,
                transactionHash: event.transactionHash,
                timestamp: (blockData === null || blockData === void 0 ? void 0 : blockData.timestamp) || 0,
                metaEvidenceId: (_a = event.args) === null || _a === void 0 ? void 0 : _a._metaEvidenceID.toString(),
                evidence: (_b = event.args) === null || _b === void 0 ? void 0 : _b._evidence,
            });
        };
        // Use provided map function or default to MetaEvidence mapping
        const mapper = mapFunction || defaultMapFunction;
        for (const event of events) {
            try {
                // Get block data
                const blockData = await event.getBlock();
                results.push(mapper(event, blockData, transactionId));
            }
            catch (error) {
                console.warn(`Error processing event at block ${event.blockNumber}:`, error);
                // Add with minimal information if block data retrieval fails
                results.push(mapper(event, null, transactionId));
            }
        }
        return results;
    }
    /**
     * Gets dispute creation events
     * @param transactionId The ID of the transaction
     * @param options Optional filter options including fromBlock and toBlock
     * @returns An array of dispute creation events
     */
    async getDisputeEvents(transactionId, options) {
        return this.searchEventsWithExtendedRange(transactionId || "unknown", "Dispute", this.escrowContract.filters.Dispute(), (event, blockData, txId) => {
            var _a, _b, _c, _d;
            return ({
                transactionId: txId,
                blockNumber: event.blockNumber,
                transactionHash: event.transactionHash,
                timestamp: (blockData === null || blockData === void 0 ? void 0 : blockData.timestamp) || 0,
                disputeId: (_a = event.args) === null || _a === void 0 ? void 0 : _a._disputeID.toString(),
                arbitrator: (_b = event.args) === null || _b === void 0 ? void 0 : _b._arbitrator.toString(),
                metaEvidenceId: (_c = event.args) === null || _c === void 0 ? void 0 : _c._metaEvidenceID.toString(),
                evidenceGroupId: (_d = event.args) === null || _d === void 0 ? void 0 : _d._evidenceGroupID.toString()
            });
        }, options);
    }
    /**
     * Gets evidence submission events
     * @param transactionId The ID of the transaction
     * @param options Optional filter options including fromBlock and toBlock
     * @returns An array of evidence submission events
     */
    async getEvidenceEvents(transactionId, options) {
        return this.searchEventsWithExtendedRange(transactionId || "unknown", "Evidence", this.escrowContract.filters.Evidence(), (event, blockData, txId) => {
            var _a, _b, _c, _d;
            return ({
                transactionId: txId,
                blockNumber: event.blockNumber,
                transactionHash: event.transactionHash,
                timestamp: (blockData === null || blockData === void 0 ? void 0 : blockData.timestamp) || 0,
                arbitrator: (_a = event.args) === null || _a === void 0 ? void 0 : _a._arbitrator.toString(),
                evidenceGroupId: (_b = event.args) === null || _b === void 0 ? void 0 : _b._evidenceGroupID.toString(),
                party: (_c = event.args) === null || _c === void 0 ? void 0 : _c._party.toString(),
                evidence: (_d = event.args) === null || _d === void 0 ? void 0 : _d._evidence
            });
        }, options);
    }
    /**
     * Gets payment events
     * @param transactionId The ID of the transaction
     * @param options Optional filter options including fromBlock and toBlock
     * @returns An array of payment events
     */
    async getPaymentEvents(transactionId, options) {
        return this.searchEventsWithExtendedRange(transactionId || "unknown", "Payment", this.escrowContract.filters.Payment(), (event, blockData, txId) => {
            var _a, _b, _c;
            return ({
                transactionId: ((_a = event.args) === null || _a === void 0 ? void 0 : _a._transactionID.toString()) || txId,
                blockNumber: event.blockNumber,
                transactionHash: event.transactionHash,
                timestamp: (blockData === null || blockData === void 0 ? void 0 : blockData.timestamp) || 0,
                amount: (_b = event.args) === null || _b === void 0 ? void 0 : _b._amount.toString(),
                party: (_c = event.args) === null || _c === void 0 ? void 0 : _c._party.toString()
            });
        }, options);
    }
    /**
     * Gets HasToPayFee events for a transaction
     * @param transactionId The ID of the transaction
     * @param options Optional filter options including fromBlock and toBlock
     * @returns Array of HasToPayFee events
     */
    async getHasToPayFeeEvents(transactionId, options) {
        return this.searchEventsWithExtendedRange(transactionId || "unknown", "HasToPayFee", this.escrowContract.filters.HasToPayFee(), (event, blockData, txId) => {
            var _a, _b;
            return ({
                transactionId: ((_a = event.args) === null || _a === void 0 ? void 0 : _a._transactionID.toString()) || txId,
                blockNumber: event.blockNumber,
                transactionHash: event.transactionHash,
                timestamp: (blockData === null || blockData === void 0 ? void 0 : blockData.timestamp) || 0,
                party: (_b = event.args) === null || _b === void 0 ? void 0 : _b._party
            });
        }, options);
    }
    /**
     * Gets ruling events
     * @param transactionId The ID of the transaction
     * @param options Optional filter options including fromBlock and toBlock
     * @returns An array of ruling events
     */
    async getRulingEvents(transactionId, options) {
        return this.searchEventsWithExtendedRange(transactionId || "unknown", "Ruling", this.escrowContract.filters.Ruling(), (event, blockData, txId) => {
            var _a, _b, _c;
            return ({
                transactionId: txId,
                blockNumber: event.blockNumber,
                transactionHash: event.transactionHash,
                timestamp: (blockData === null || blockData === void 0 ? void 0 : blockData.timestamp) || 0,
                arbitrator: (_a = event.args) === null || _a === void 0 ? void 0 : _a._arbitrator.toString(),
                disputeId: (_b = event.args) === null || _b === void 0 ? void 0 : _b._disputeID.toString(),
                ruling: (_c = event.args) === null || _c === void 0 ? void 0 : _c._ruling.toNumber()
            });
        }, options);
    }
}
